/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2005 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "stdafx.h"
#include "TaskbarButton.h"
#include "Taskbar.h"
#include "WindowUtilities.h"
#include "ButtonTexture.h"
#include "ButtonIcon.h"
#include "ButtonFont.h"

//Grouping
#define	X_GROUP_MINIMIZE	10
#define	X_GROUP_RESTORE		11
#define	X_GROUP_MAXIMIZE	12
#define	X_GROUP_CLOSE		13

#define	X_GROUP_SEPARATOR	66

TaskbarButton::TaskbarButton( Taskbar *actualtaskbar, HWND hWnd ):taskbar(NULL), buttonx(0), buttony(0), buttonheight(0), buttonwidth(0)
{
	hwndgrouplist.clear();
	
	useHover = false;
	flashNow = false;

	currentState = STATE_NORMAL;

	taskbar = actualtaskbar;

	hButtonTask = hWnd;

	classname = GetTaskFileName( hWnd );

	hwndgrouplist.insert(hwndgrouplist.end(), hWnd);

	position = -1;

	icon = NULL;

	if(taskbar->showText || taskbar->showToolTip)
		caption = GetWindowCaption();
	else
		caption = "";
	
	if (taskbar->showToolTip)
		tooltip = (caption + "\n" + classname);
}

TaskbarButton::~TaskbarButton()
{
	hwndgrouplist.clear();

	if (icon)
		DestroyIcon(icon);
}

BOOL TaskbarButton::HasPoint( int xP, int yP )
{
	if ( xP < (buttonx + buttonwidth) && xP >= buttonx )
	{
		if ( yP < (buttony + buttonheight) && yP >= buttony )
			return TRUE;
	}

	return FALSE;
}

BOOL TaskbarButton::onUpdate(bool needRedraw)
{
	if (taskbar->showText || taskbar->showToolTip)
	{
		string temp1 = caption;
		string temp2 = GetWindowCaption();

		if ( stricmp( temp1.c_str(), temp2.c_str() ) != NULL && !temp2.empty() )
		{
			caption = temp2;
			if (taskbar->showToolTip && hwndgrouplist.size() > 1)
			{
				temp2.append( "\nGrouped (" + IntToString(hwndgrouplist.size()) + ")\n" + classname );
				tooltip = temp2;
			}
			else if (taskbar->showToolTip)
			{	
				tooltip = (temp2 + "\n" + classname);
			}
			if (taskbar->showText)
				needRedraw = true;
		}
	}

	if (taskbar->displayButtonType >= 0)
		needRedraw = true;

	return needRedraw;	
}

string TaskbarButton::GetWindowCaption( HWND check_hTask )
{
	if (check_hTask == NULL)
		check_hTask = hButtonTask;
	
	return GetTaskWindowText(check_hTask);
}

void TaskbarButton::onButtonClick( int x, int y, int action, bool minimize )
{
	switch (action)
	{
		case 1:
			
			Select(minimize);
		    break;

		case 2:

			Menu();
		    break;

		case 3:

			Close(hButtonTask);
		    break;

		case 4:

			OnTopToggle(hButtonTask);
			break;
	}
}

void TaskbarButton::Select(bool minimize)
{
	if( flashNow )
	{
		SetFlashing(FALSE);
		currentState = getCurrentState();
	}

	if (hwndgrouplist.size() > 1 && taskbar->useGrouping)
	{
		GroupMenu();
		return;
	}

	if ( minimize )
	{
		if (!taskbar->noMinimizeOnClick)
		{
			MinimizeWindow(hButtonTask);
		}
		else
		{
			SelectWindow(hButtonTask);
			if ( !taskbar->taskbarSelect.empty() )
				LSExecute(taskbar->gethWnd(), taskbar->taskbarSelect.c_str(), SW_SHOWNORMAL);
		}
	}
	else
	{
		switch(currentState)
		{
			case STATE_SELECTED:

				if (!taskbar->noMinimizeOnClick)
				{
					MinimizeWindow(hButtonTask);
				}
				else
				{
					SelectWindow(hButtonTask);
					if ( !taskbar->taskbarSelect.empty() )
						LSExecute(taskbar->gethWnd(), taskbar->taskbarSelect.c_str(), SW_SHOWNORMAL);
				}

			break;

			case STATE_MINIMIZED:

				SelectWindow(hButtonTask, true);

			break;

			default:

				SelectWindow(hButtonTask);
				if ( !taskbar->taskbarSelect.empty() )
					LSExecute(taskbar->gethWnd(), taskbar->taskbarSelect.c_str(), SW_SHOWNORMAL);

			break;
		}
	}
}

void TaskbarButton::Menu()
{
	HMENU systemMenu;

	systemMenu = GetSystemMenu( hButtonTask, FALSE );

	WINDOWPLACEMENT wp;
	wp.length = sizeof(WINDOWPLACEMENT);
	GetWindowPlacement( hButtonTask, &wp );

	// restore is enabled only if minimized or maximized (not normal)
	EnableMenuItem( systemMenu, SC_RESTORE, MF_BYCOMMAND | (wp.showCmd != SW_SHOWNORMAL ? MF_ENABLED : MF_GRAYED) );

	// move is enabled only if normal
	EnableMenuItem( systemMenu, SC_MOVE, MF_BYCOMMAND | (wp.showCmd == SW_SHOWNORMAL ? MF_ENABLED : MF_GRAYED) );

	// size is enabled only if normal
	EnableMenuItem( systemMenu, SC_SIZE, MF_BYCOMMAND | (wp.showCmd == SW_SHOWNORMAL ? MF_ENABLED : MF_GRAYED) );

	// minimize is enabled only if not minimized
	EnableMenuItem( systemMenu, SC_MOVE, MF_BYCOMMAND | (wp.showCmd != SW_SHOWMINIMIZED ? MF_ENABLED : MF_GRAYED) );

	// maximize is enabled only if not maximized
	EnableMenuItem( systemMenu, SC_MOVE, MF_BYCOMMAND | (wp.showCmd != SW_SHOWMAXIMIZED ? MF_ENABLED : MF_GRAYED) );

	int counter = 0;

	if (hwndgrouplist.size() > 1 && taskbar->useGrouping)
	{	
		AppendMenu( systemMenu, MF_SEPARATOR, X_GROUP_SEPARATOR, "" );

		AppendMenu( systemMenu, MF_STRING, X_GROUP_MINIMIZE, "Minimize All In Group" );
		AppendMenu( systemMenu, MF_STRING, X_GROUP_RESTORE, "Restore All In Group" );
		AppendMenu( systemMenu, MF_STRING, X_GROUP_MAXIMIZE, "Maximize All In Group" );
		AppendMenu( systemMenu, MF_STRING, X_GROUP_CLOSE, "Close All In Group" );

		AppendMenu( systemMenu, MF_SEPARATOR, X_GROUP_SEPARATOR, "" );

		counter = 1;
		for(hWndGroupListIterator hglit = hwndgrouplist.begin(); hglit != hwndgrouplist.end(); ++hglit)
		{
			AppendMenu( systemMenu, MF_STRING , counter, ("Close " + GetWindowCaption(*hglit)).c_str() );
			counter++;
		}
		counter--;
		//END Grouping
	}

	// let application modify menu
	PostMessage( hButtonTask, WM_INITMENUPOPUP, (WPARAM) systemMenu, MAKELPARAM( 0, TRUE ) );
	PostMessage( hButtonTask, WM_INITMENU, (WPARAM) systemMenu, 0 );

	// display the menu
	POINT pt;
	GetCursorPos( &pt );

	int command = TrackPopupMenu( systemMenu,
	                              TPM_RETURNCMD | TPM_RIGHTBUTTON,
	                              pt.x, pt.y, 0,
	                              taskbar->gethWnd(),
	                              NULL );

	if (hwndgrouplist.size() > 1 && taskbar->useGrouping)
	{
		DeleteMenu( systemMenu, X_GROUP_SEPARATOR, MF_SEPARATOR );

		DeleteMenu( systemMenu, X_GROUP_SEPARATOR, MF_BYCOMMAND );
		DeleteMenu( systemMenu, X_GROUP_MINIMIZE, MF_BYCOMMAND );
		DeleteMenu( systemMenu, X_GROUP_RESTORE, MF_BYCOMMAND );
		DeleteMenu( systemMenu, X_GROUP_MAXIMIZE, MF_BYCOMMAND );
		DeleteMenu( systemMenu, X_GROUP_CLOSE, MF_BYCOMMAND );

		DeleteMenu( systemMenu, X_GROUP_SEPARATOR, MF_SEPARATOR );

		while ( counter >= 1 )
		{
			DeleteMenu( systemMenu, counter, MF_BYCOMMAND );
			counter--;
		}
	}

	if ( command )
	{
		if (hwndgrouplist.size() > 1 && taskbar->useGrouping)
		{
			switch( command )
			{
				case X_GROUP_MINIMIZE:
				{
					for ( hWndGroupListIterator hglit = hwndgrouplist.begin(); hglit != hwndgrouplist.end(); ++hglit )
						MinimizeWindow( (*hglit) );
					
					return;
				}

				case X_GROUP_RESTORE:
				{
					for ( hWndGroupListIterator hglit = hwndgrouplist.begin(); hglit != hwndgrouplist.end(); ++hglit )
						RestoreWindow( (*hglit) );

					return;
				}

				case X_GROUP_MAXIMIZE:
				{
					for ( hWndGroupListIterator hglit = hwndgrouplist.begin(); hglit != hwndgrouplist.end(); ++hglit )
						MaximizeWindow( (*hglit) );

					return;
				}

				case X_GROUP_CLOSE:
				{
					for ( hWndGroupListIterator hglit = hwndgrouplist.begin(); hglit != hwndgrouplist.end(); ++hglit )
						Close( (*hglit) );
						
					return;
				}
			}

			hWndGroupListIterator hglit = hwndgrouplist.begin();
			counter = 1; 
			while (counter < command)
			{
				hglit++;
				counter++;
			}

			if ( command <= hwndgrouplist.size() )
			{
				Close( (*hglit) );
				hButtonTask = (*hglit);
				return;
			}
		}
		
		if ( command != SC_MINIMIZE && command != SC_CLOSE )
		{
			SelectWindow(hButtonTask);
			if ( !taskbar->taskbarSelect.empty() )
				LSExecute(taskbar->gethWnd(), taskbar->taskbarSelect.c_str(), SW_SHOWNORMAL);
		}

		PostMessage( hButtonTask, WM_SYSCOMMAND, (WPARAM) command, MAKELPARAM( pt.x, pt.y ) );
	}
}

void TaskbarButton::GroupMenu()
{
	HMENU GroupMenu = CreatePopupMenu();

	int counter = 1;
	for(hWndGroupListIterator hglit = hwndgrouplist.begin(); hglit != hwndgrouplist.end(); ++hglit)
	{
		AppendMenu( GroupMenu, MF_STRING, counter, GetWindowCaption((*hglit)).c_str() );
		counter++;
	}

	// display the menu
	POINT pt;
	GetCursorPos( &pt );

	int command = TrackPopupMenu( GroupMenu,
	                              TPM_RETURNCMD | TPM_RIGHTBUTTON,
	                              pt.x, pt.y, 0,
	                              taskbar->gethWnd(),
	                              NULL );

	if ( command )
	{
		hWndGroupListIterator hglit = hwndgrouplist.begin();
		counter = 1; 
		while (counter < command)
		{
			hglit++;
			counter++;
		}

		if ( counter <= hwndgrouplist.size() )
		{
			SelectWindow( (*hglit) );
			if ( !taskbar->taskbarSelect.empty() )
				LSExecute(taskbar->gethWnd(), taskbar->taskbarSelect.c_str(), SW_SHOWNORMAL);
			hButtonTask = (*hglit);
		}
	}
	
	if (GroupMenu)
		DestroyMenu(GroupMenu);
}

int TaskbarButton::getCurrentState()
{
	return GetState(hButtonTask);
}

void TaskbarButton::setButtonLayout( int x, int y, int width, int height )
{
	buttonx = x;
	buttony = y;
	buttonwidth = width;	
	buttonheight = height;
}

void TaskbarButton::onPaint( HDC destDC )
{
	//Update Icon - Sure is Sure
	if ( taskbar->showIcon )
	{
		if (icon)
		{
			DestroyIcon(icon);
			icon = NULL;
		}

		if ( GetIconOnCustomList( hButtonTask, icon, taskbar->CustomIconList, taskbar->useBigIcon) )
			icon = (HICON)CopyImage((HANDLE)icon, IMAGE_ICON, 0, 0, 0);
	}

	if (!flashNow)
	{
		currentState = getCurrentState();

		if( STATE_SELECTED == currentState )
		{
			taskbar->lastActive = hButtonTask;
		}
	}

	if ( flashNow && STATE_FLASHING == currentState )
		currentState = STATE_NORMAL;
	else if ( flashNow )
		currentState = STATE_FLASHING;

	//Button Background Painting with Icon and Text 
	if(currentState == STATE_FLASHING)
	{
		taskbar->buttontexture->apply(destDC, buttonx, buttony, buttonwidth, buttonheight, 3);
		
		if(taskbar->showIcon)
		{
			taskbar->buttonicon->apply(destDC, buttonx, buttony, buttonwidth, buttonheight, 3, 3, icon, taskbar->isAlphaMap());
		}
		if (taskbar->showText)
		{	
			taskbar->buttonfont->apply(destDC, buttonx, buttony, buttonwidth, buttonheight, caption, 3, 3);
		}
	}
	else if (taskbar->useGrouping && hwndgrouplist.size() > 1 && !useHover )
	{
		int tmpButtonTextureType = 5;
	
		if (currentState == STATE_SELECTED)
			tmpButtonTextureType = 1;
		else if (currentState == STATE_MINIMIZED)
			tmpButtonTextureType = 2;
		else
			tmpButtonTextureType = 0;
		
		taskbar->buttontexture->apply(destDC, buttonx, buttony, buttonwidth, buttonheight, (taskbar->buttontexture->useGroupedBG() ? 5 : tmpButtonTextureType) );

		if(taskbar->showIcon)
		{
			taskbar->buttonicon->apply(destDC, buttonx, buttony, buttonwidth, buttonheight, 5, tmpButtonTextureType, icon, taskbar->isAlphaMap());
		}
		if (taskbar->showText)
		{
			CString text = caption.c_str();
		
			int found = text.Find(" - ");
			if ( found != -1 && text.Find(" - ", found+3) != -1)
				text = text.Mid(found+3, text.Find(" - ", found+3)-found-3);
			else if ( found != -1 )
				text = text.Left(text.Find(" - "));
			text += " ( ";
			text += IntToString( hwndgrouplist.size() ).c_str();
			text += " )";

			taskbar->buttonfont->apply(destDC, buttonx, buttony, buttonwidth, buttonheight, (string) text, 5, tmpButtonTextureType);
		}
	}
	else if (useHover)
	{
		int tmpButtonTextureType = 4;
		
		if (currentState == STATE_SELECTED)
			tmpButtonTextureType = 1;
		else if (currentState == STATE_MINIMIZED)
			tmpButtonTextureType = 2;
		else
			tmpButtonTextureType = 0;
	
		taskbar->buttontexture->apply(destDC, buttonx, buttony, buttonwidth, buttonheight, (taskbar->buttontexture->useHoverBG() ? 4: tmpButtonTextureType));

		if(taskbar->showIcon)
		{
			taskbar->buttonicon->apply(destDC, buttonx, buttony, buttonwidth, buttonheight, 4, tmpButtonTextureType, icon, taskbar->isAlphaMap());
		}
		if (taskbar->showText)
		{
			if (taskbar->useGrouping && hwndgrouplist.size() > 1)
			{
				CString text = caption.c_str();
			
				int found = text.Find(" - ");
				if ( found != -1 && text.Find(" - ", found+3) != -1)
					text = text.Mid(found+3, text.Find(" - ", found+3)-found-3);
				else if ( found != -1 )
					text = text.Left(text.Find(" - "));
				text += " ( ";
				text += IntToString( hwndgrouplist.size() ).c_str();
				text += " )";

				taskbar->buttonfont->apply(destDC, buttonx, buttony, buttonwidth, buttonheight, (string) text, 4, tmpButtonTextureType);
			}
			else
			{
				taskbar->buttonfont->apply(destDC, buttonx, buttony, buttonwidth, buttonheight, caption, 4, tmpButtonTextureType);
			}
		}
	}
	else
	{
		if (currentState == STATE_SELECTED)
		{
			taskbar->buttontexture->apply(destDC, buttonx, buttony, buttonwidth, buttonheight, 1);
			
			if(taskbar->showIcon)
			{
				taskbar->buttonicon->apply(destDC, buttonx, buttony, buttonwidth, buttonheight, 1, 1, icon, taskbar->isAlphaMap());
			}
			if (taskbar->showText)
			{
				taskbar->buttonfont->apply(destDC, buttonx, buttony, buttonwidth, buttonheight, caption, 1, 1);
			}
		}
		else if (currentState == STATE_MINIMIZED)
		{
			taskbar->buttontexture->apply(destDC, buttonx, buttony, buttonwidth, buttonheight, 2);
			
			if(taskbar->showIcon)
			{
				taskbar->buttonicon->apply(destDC, buttonx, buttony, buttonwidth, buttonheight, 2, 2, icon, taskbar->isAlphaMap());
			}
			if (taskbar->showText)
			{
				taskbar->buttonfont->apply(destDC, buttonx, buttony, buttonwidth, buttonheight, caption, 2, 2);
			}
		}
		else
		{
			taskbar->buttontexture->apply(destDC, buttonx, buttony, buttonwidth, buttonheight, 0);
		
			if(taskbar->showIcon)
			{
				taskbar->buttonicon->apply(destDC, buttonx, buttony, buttonwidth, buttonheight, 0, 0, icon, taskbar->isAlphaMap());
			}
			if (taskbar->showText)
			{
				taskbar->buttonfont->apply(destDC, buttonx, buttony, buttonwidth, buttonheight, caption, 0, 0);
			}
		}
	}
}
